<?php
/**
 * User: nester_a
 * Date: 02.07.14
 */

namespace LMS\Libraries;

class ServiceLocator
{
	private static $namespaces = array();
	private static $activated = false;
	private static $needToAddS = array(
		'View', 'Model'
	);
	private static $services = array();
	private static $mocks = array();
	private static $serviceConfig = array();

	/**
	 *
	 */
	public static function activate()
	{
		if(self::$activated){
			return;
		}else{
			self::$activated = true;
			spl_autoload_register(function ($class) {
				ServiceLocator::autoloadCallback($class);
			});
		}
	}

	/**
	 * @param $namespace
	 * @param $path
	 */
	public static function registerNamespace($namespace, $path)
	{
		self::$namespaces[$namespace] = $path;
	}

	/**
	 * @param $serviceName
	 * @param bool $isNew
	 * @param array $arguments
	 * @return mixed
	 * @throws \InvalidArgumentException
	 */
	public static function get($serviceName, $isNew = false, $arguments=array())
	{
		if(empty(self::$serviceConfig) && empty(self::$mocks)){
			throw new \InvalidArgumentException('Trying to load "'.$serviceName.'". Please load service config first or set mock class');
		}
		return self::getService(self::$serviceConfig, $serviceName, $isNew, $arguments);
	}

	/**
	 * @param $configFileAlias
	 * @param bool $prepend
	 */
	public static function load($configFileAlias, $prepend = false)
	{
		self::$serviceConfig = $prepend == true ? array_merge(self::loadConfig($configFileAlias), self::$serviceConfig) : array_merge(self::$serviceConfig, self::loadConfig($configFileAlias));
	}

	/**
	 * Set mocks for testing purposes
	 *
	 * @param array $mocks
	 * @throws \InvalidArgumentException
	 */
	public static function setMocks($mocks)
	{
		if(!is_array($mocks)){
			throw new \InvalidArgumentException('Mocks should be array in format "serviceAlias" => "mockClass"');
		}

		self::$mocks = array_merge(self::$mocks, $mocks);
	}

	/**
	 * @param $class
	 */
	public static function autoloadCallback($class)
	{
		$parsedClassName = self::parseClassName($class);

		if(count($parsedClassName) < 3){
			return;
		}

		if(in_array($parsedClassName[count($parsedClassName)-2], self::$needToAddS)){
			$parsedClassName[count($parsedClassName)-2] = $parsedClassName[count($parsedClassName)-2] . 's';
		}

		$pathToFile = self::buildPath($parsedClassName);
		$pathToFileForShortenClassName = self::buildPathForShortenClassName($parsedClassName);

		if(!isset(self::$namespaces[$parsedClassName[0]])){
			return;
		}

		$namespacePath = self::$namespaces[$parsedClassName[0]];
		$found = false;
		if(file_exists($namespacePath . $pathToFile)){
			include $namespacePath . $pathToFile;
			$found = true;
		}

		if(!$found){
			if(file_exists($namespacePath . $pathToFileForShortenClassName)){
				include $namespacePath . $pathToFileForShortenClassName;
			}
		}
	}


	/**
	 * @param $serviceAlias
	 * @param array $arguments
	 * @return mixed
	 */
	private static function getExternalService($serviceAlias, $arguments=array())
	{
		$parsedAlias = self::parseAlias($serviceAlias);
		$serviceName = array_pop($parsedAlias);
		$configAlias = implode('.', $parsedAlias);
		$config = self::loadConfig($configAlias);
		return self::getService($config, $serviceName, $arguments);

	}

	/**
	 * @param $config
	 * @param $serviceName
	 * @param bool $isNew
	 * @param array $arguments
	 * @return mixed
	 * @throws \Exception
	 */
	private static function getService($config, $serviceName, $isNew = false, $arguments = array())
	{
		$hash = sha1($serviceName . serialize($arguments));

		if(isset(self::$mocks[$serviceName])){
			return self::$mocks[$serviceName];
		}

		// return service from cache
		if(isset(self::$services[$hash]) && !$isNew){
			return self::$services[$hash];
		}

		if(!isset($config[$serviceName])){
			throw new \Exception('There is no service with name "'.$serviceName.'"defined');
		}

		$service = $config[$serviceName];

		// if link to other service then return it
		if(is_string($service) && $service[0] == '@'){
			return self::getExternalService(substr($service, 1));
		}

		if(isset($service['class'])){
			$reflectionClass = new \ReflectionClass($service['class']);

			if(isset($service['args'])){
				foreach($service['args'] as &$arg){
					if(is_array($arg)){
						foreach($arg as &$value){
							if($value[0] == '@'){
								$value = self::getExternalService(substr($value, 1));
							}
						}
					}else{
						if($arg[0] == '@'){
							$arg = self::getExternalService(substr($arg, 1));
						}
					}
				}
				$arguments = array_merge($service['args'], $arguments);
			}

			if(!empty($arguments)){
				self::$services[$hash] = $reflectionClass->newInstanceArgs($arguments);
			}else{
				self::$services[$hash] = $reflectionClass->newInstance();
			}
			return self::$services[$hash];
		}

		if(isset($service['factory'])){

			if(isset($service['args'])){
				$arguments = array_merge($service['args'], $arguments);
			}

			if(!empty($arguments)){
				self::$services[$hash] = call_user_func_array($service['factory'].'::'.$service['method'], $arguments);
			}else{
				self::$services[$hash] = call_user_func($service['factory'].'::'.$service['method']);
			}

			return self::$services[$hash];
		}

		throw new \Exception('Provided wrong config format for '.$serviceName);
	}

	/**
	 * @param $versionNumber
	 * @return bool
	 */
	private static function isCompatible($versionNumber)
	{
		jimport('joomla.version');
		$version = new \JVersion();
		return $version->isCompatible($versionNumber);
	}

	/**
	 * @param $configFileAlias
	 * @return array
	 */
	private static function loadConfig($configFileAlias)
	{
		$servicesPath = array(
			'tool' => __DIR__ . '/../tools',
			'lib' => __DIR__ ,
			'widget' => __DIR__ . '/../widgets',
			'test' => __DIR__ . '/../tests'
		);

		$parsedAlias = self::parseAlias($configFileAlias);
		$path = $servicesPath[$parsedAlias[0]];

		$pathToConfig = $path;
		$lastIndex = count($parsedAlias)-1;
		for($i=1; $i<$lastIndex; $i++){
			$pathToConfig .= '/' . $parsedAlias[$i];
		}

		$pathToConfig .= '/' . $parsedAlias[$lastIndex] . '/' . $parsedAlias[$lastIndex] .'.json';

		if(!file_exists($pathToConfig)){
			return array();
		}

		$config = json_decode(@file_get_contents($pathToConfig), true);
		
		if(is_null($config)) {
			throw new \Exception('Provided wrong config format for '.$pathToConfig);
		}
		
		$mainConfig = $config['*'];
		array_shift($config);

		$versionedConfig = array();
		foreach($config as $version => $versionConfig){
			if(self::isCompatible($version)){
				$versionedConfig = array_merge($versionedConfig, $versionConfig);
			}
		}

		return array_merge($mainConfig, $versionedConfig);
	}

	/**
	 * @param $alias
	 * @return array
	 */
	private static function parseAlias($alias)
	{
		return explode('.', $alias);
	}

	/**
	 * @param $className
	 * @return array
	 */
	private static function parseClassName($className)
	{
		return explode('\\', $className);
	}

	/**
	 * @param $parsedName
	 * @return string
	 */
	private static function buildPath($parsedName)
	{
		$path = '';
		$count = count($parsedName);
		for($i=1; $i<$count; $i++){
			$path .= '/'.strtolower($parsedName[$i]);
		}

		return $path . '.php';
	}

	/**
	 * @param $parsedName
	 * @return string
	 */
	private static function buildPathForShortenClassName($parsedName)
	{
		$path = '';
		$count = count($parsedName);
		for($i=1; $i<$count; $i++){
			$path .= '/'.strtolower($parsedName[$i]);
		}


		return $path . '/' . strtolower($parsedName[count($parsedName)-1]) . '.php';
	}

}